#include <stdio.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include "guiutils.h"
#include "grabopts.h"

#include "images/icon_camera_20x20.xpm"
#include "images/icon_cancel_20x20.xpm"
#include "grabopts.xpm"


static gint close_cb(GtkWidget *widget, GdkEvent *event, gpointer data);
static void destroy_cb(GtkObject *object, gpointer data);
static void grab_btn_cb(GtkWidget *widget, gpointer data);
static void cancel_btn_cb(GtkWidget *widget, gpointer data);

static void build_std_dialog(
        gint width, gint height, const gchar *title,
        GtkWidget *ref_toplevel,
        GtkWidget **toplevel_rtn, GtkWidget **client_vbox_rtn,
        GtkAccelGroup **accelgrp_rtn
);


gint iv_query_grab_opts(
        GtkWidget *ref_toplevel,
        glong *delay,           /* In milliseconds. */
	gbool *hide_windows,
        gbool *include_decorations
);


#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)     (((x) < 0) ? ((x) * -1) : (x))


#define BTN_WIDTH	(100 + (2 * 3))
#define BTN_HEIGHT	(30 + (2 * 3))


/* Instructions for grabbing displayed on grab dialog. */
#define GRAB_INSTRUCTIONS_MESG	"\
After grab delay; press Button1 to grab a single\n\
window, press Button2 to grab a rectangular area,\n\
or press Button3 to abort grab. To grab the entire\n\
desktop press Button1 over the root window."


/* Tooltip messages. */
#define GRAB_TTM_DELAY		"Delay in seconds before GUI is locked\
 for grabbing"
#define GRAB_TTM_HIDE_WINDOWS	"Hide this application's windows during\
 grab"
#define GRAB_TTM_INCLUDE_DECORATIONS	"Include window manager decoration\
 windows when grabbing a single window"


static gbool got_response;


/*
 *	Closes the given toplevel window given by data and breaks
 *	the highest gtk_main level.
 */
static gint close_cb(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	cancel_btn_cb(widget, data);
	return(TRUE);
}

/*
 *	Destroy callback.
 */
static void destroy_cb(GtkObject *object, gpointer data)
{
	return;
}

/*
 *	Save button callback.
 */
static void grab_btn_cb(GtkWidget *widget, gpointer data)
{
        GtkWidget *toplevel = (GtkWidget *)data;
        if(toplevel == NULL)
            return;

	got_response = TRUE;

        gtk_main_quit();
}

/*
 *	Cancel button callback.
 */
static void cancel_btn_cb(GtkWidget *widget, gpointer data)
{
        GtkWidget *toplevel = (GtkWidget *)data;
        if(toplevel == NULL)
            return;

        got_response = FALSE;

        gtk_main_quit();
}

/*
 *	Builds a standard toplevel dialog and returns the pointers to
 *	the created widgets.
 */
static void build_std_dialog(
	gint width, gint height, const gchar *title,
	GtkWidget *ref_toplevel,
	GtkWidget **toplevel_rtn, GtkWidget **client_vbox_rtn,
	GtkAccelGroup **accelgrp_rtn
)
{
	gint	bw = BTN_WIDTH,
		bh = BTN_HEIGHT,
		border_major = 5;
	GdkWindow *window;
	GtkAccelGroup *accelgrp;
        GtkWidget *w, *parent, *parent2;
        GtkWidget *toplevel, *main_vbox;


        /* Create keyboard accelerator group. */
        accelgrp = gtk_accel_group_new();
	if(accelgrp_rtn != NULL)
	    (*accelgrp_rtn) = accelgrp;

        /* Create toplevel. */
        toplevel = parent = w = gtk_window_new(GTK_WINDOW_DIALOG);
        gtk_widget_set_usize(w, width, height);
        gtk_widget_realize(w);
        gtk_window_set_title(GTK_WINDOW(w), title);
        window = w->window;
        if(window != NULL)
        {
            gdk_window_set_decorations(
                window,
                GDK_DECOR_TITLE | GDK_DECOR_MENU | GDK_DECOR_MINIMIZE
            );
            gdk_window_set_functions(
                window,
                GDK_FUNC_MOVE | GDK_FUNC_MINIMIZE | GDK_FUNC_CLOSE
            );
	    GUISetWMIcon(window, (u_int8_t **)grabopts_xpm);
        }
        gtk_signal_connect(
            GTK_OBJECT(w), "delete_event",
            GTK_SIGNAL_FUNC(close_cb),
            (gpointer)toplevel
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "destroy",
            GTK_SIGNAL_FUNC(destroy_cb),
            (gpointer)toplevel
        );
        gtk_container_border_width(GTK_CONTAINER(w), 0);
        gtk_accel_group_attach(accelgrp, GTK_OBJECT(w));
        if((ref_toplevel != NULL) ? GTK_IS_WINDOW(GTK_OBJECT(ref_toplevel)) : 0)
        {
            gtk_window_set_modal(GTK_WINDOW(toplevel), TRUE);
            gtk_window_set_transient_for(
                GTK_WINDOW(toplevel), GTK_WINDOW(ref_toplevel)
            );
        }
	if(toplevel_rtn != NULL)
	    (*toplevel_rtn) = toplevel;

        main_vbox = w = gtk_vbox_new(FALSE, 0);
        gtk_container_add(GTK_CONTAINER(parent), w);
        gtk_widget_show(w);
        parent = w;

        w = gtk_vbox_new(FALSE, border_major);
        gtk_container_border_width(GTK_CONTAINER(w), border_major);
        gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
        gtk_widget_show(w);
        parent2 = w;
	if(client_vbox_rtn != NULL)
	    (*client_vbox_rtn) = w;

	w = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(main_vbox), w, FALSE, FALSE, 0);
        gtk_widget_show(w);

        /* Buttons. */
        w = gtk_hbox_new(TRUE, 0);
        gtk_box_pack_start(GTK_BOX(main_vbox), w, FALSE, FALSE, border_major);
        gtk_widget_show(w);
        parent2 = w;

        /* Grab button. */
        w = GUIButtonPixmapLabelH(
            (u_int8_t **)icon_camera_20x20_xpm, "Grab", NULL
        );
        gtk_widget_set_usize(w, bw, bh);
        GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT | GTK_HAS_DEFAULT);
        gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(w), "clicked",
            GTK_SIGNAL_FUNC(grab_btn_cb),
            (gpointer)toplevel
        );
        gtk_accel_group_add(
            accelgrp, GDK_Return, 0, GTK_ACCEL_VISIBLE,
            GTK_OBJECT(w), "clicked"
        );
        gtk_accel_group_add(
            accelgrp, GDK_3270_Enter, 0, GTK_ACCEL_VISIBLE,
            GTK_OBJECT(w), "clicked"
        );
        gtk_accel_group_add(
            accelgrp, GDK_KP_Enter, 0, GTK_ACCEL_VISIBLE,
            GTK_OBJECT(w), "clicked"
        );
        gtk_accel_group_add(
            accelgrp, GDK_ISO_Enter, 0, GTK_ACCEL_VISIBLE,
            GTK_OBJECT(w), "clicked"
        );
        gtk_widget_show(w);

        /* Cancel button. */
        w = GUIButtonPixmapLabelH(
            (u_int8_t **)icon_cancel_20x20_xpm, "Cancel", NULL
        );
        gtk_widget_set_usize(w, bw, bh);
        GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
        gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(w), "clicked",
            GTK_SIGNAL_FUNC(cancel_btn_cb),
            (gpointer)toplevel
        );
        gtk_accel_group_add(
            accelgrp, GDK_Escape, 0, GTK_ACCEL_VISIBLE,
            GTK_OBJECT(w), "clicked"
        );
        gtk_widget_show(w);
}


/*
 *	Standard grab options querying.
 *
 *      Blocks execution untill response is recieved.
 *      Returns TRUE if values were fetched or FALSE if user canceled.
 */
gint iv_query_grab_opts(
        GtkWidget *ref_toplevel,
	glong *delay,		/* In milliseconds. */
	gbool *hide_windows,
        gbool *include_decorations
)
{
        gint    width = 320,
                height = -1,
                border_minor = 2;
        GtkAccelGroup *accelgrp = NULL;
        GtkWidget *toplevel = NULL, *client_vbox = NULL;
	GtkWidget *delay_spin = NULL;
        GtkWidget *include_decorations_toggle = NULL;
	GtkWidget *hide_windows_toggle = NULL;


        /* Reset global got response flag. */
        got_response = FALSE;

        /* Create dialog base widgets. */
        build_std_dialog(
            width, height,
            "Grab Window",
            ref_toplevel,
            &toplevel, &client_vbox, &accelgrp
        );

        /* Create grab options widgets. */
        if(client_vbox != NULL)
        {
            GtkWidget *w, *parent = client_vbox, *parent2;
	    GtkAdjustment *adj;


            /* Instructions. */
            w = gtk_hbox_new(FALSE, border_minor);
            gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
            gtk_widget_show(w);
            parent2 = w;

	    w = gtk_label_new(GRAB_INSTRUCTIONS_MESG);
	    gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
            gtk_widget_show(w);

	    /* Delay (in seconds). */
            w = gtk_hbox_new(FALSE, border_minor);
            gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
            gtk_widget_show(w);
            parent2 = w;
	    /* Label. */
	    w = gtk_label_new("Delay:");
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    /* Spin button. */
	    adj = (GtkAdjustment *)gtk_adjustment_new(
		(gfloat)((delay != NULL) ? *delay : 0) / 1000,
		0.0,
		10.0 * 60.0 * 1000.0,	/* Max 10 minutes. */
		1.0,
		5.0,
		5.0
	    );
	    delay_spin = w = gtk_spin_button_new(
		adj,
		1.0,	/* Climb rate (0.0 to 1.0). */
		0	/* Digits. */
	    );
	    gtk_widget_set_usize(w, 75, -1);
            gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    GUISetWidgetTip(w, GRAB_TTM_DELAY);
            gtk_widget_show(w);
            /* Units label. */
            w = gtk_label_new("Seconds");
            gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
            gtk_widget_show(w);

            /* Include decorations and hide windows. */
            w = gtk_hbox_new(FALSE, border_minor);
            gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
            gtk_widget_show(w);
            parent2 = w;
            /* Check button. */
            include_decorations_toggle = w =
		gtk_check_button_new_with_label("Include Decorations"
	    );
            gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
            if((include_decorations != NULL) ? *include_decorations : FALSE)
                GTK_TOGGLE_BUTTON(w)->active = TRUE;
            GUISetWidgetTip(w, GRAB_TTM_INCLUDE_DECORATIONS);
            gtk_widget_show(w);
            /* Check button. */
            hide_windows_toggle = w =
                gtk_check_button_new_with_label("Hide Windows");
            gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
            if((hide_windows != NULL) ? *hide_windows : FALSE)
                GTK_TOGGLE_BUTTON(w)->active = TRUE;
            GUISetWidgetTip(w, GRAB_TTM_HIDE_WINDOWS);
            gtk_widget_show(w);
        }


        /* Map toplevel. */
        gtk_widget_show(toplevel);


        /* Enter blocking gtk_main loop. */
        gtk_main();


        /* Get values from widgets if user clicked on save. */
        if(got_response)
        {
	    if((delay != NULL) && (delay_spin != NULL))
		(*delay) = (glong)(gtk_spin_button_get_value_as_float(
		    GTK_SPIN_BUTTON(delay_spin)) * 1000.0);

            if((include_decorations != NULL) &&
	       (include_decorations_toggle != NULL)
	    )
		(*include_decorations) = GTK_TOGGLE_BUTTON(
		    include_decorations_toggle)->active;
            if((hide_windows != NULL) &&
               (hide_windows_toggle != NULL)
            )
                (*hide_windows) = GTK_TOGGLE_BUTTON(
                    hide_windows_toggle)->active;
        }


        /* Destroy widgets and related resources. */
        gtk_window_set_modal(GTK_WINDOW(toplevel), FALSE);
        gtk_window_set_transient_for(GTK_WINDOW(toplevel), NULL);
        gtk_widget_destroy(toplevel);
        gtk_accel_group_unref(accelgrp);


        /* Return user response. */
        return(got_response);
}
